program TextOverlay ;
{ Image text overlays }
{ Copyright (c) 2005, Digital Optics Ltd, Auckland, New Zealand }
{ Version 1.0 }

{ IMPORTANT
  This module requires the VTextOverlay.dll library defined by TextOverlayLib below.
  The .v module source code and the .dll library should both be installed in the V++
  stndard Executable folder.
  This folder is usually found at C:\Program Files\Digital Optics\V++\Executable  
}

{ INSTRUCTIONS
  
  - The following commands are available:
  
      Add Text   Adds text to an image
      Add Name   Adds the image name as text on the image
      Remove     Removes text from the last image you annotated
      Options    Opens the Text Overlay options dialog box
      
  - Users should experiment with text settings until they find font, size, style,
    and other settings that suit their images and applications. If working with
    images that are frequently zoomed down on screen, choose a larger font for
    best results.
    
  - If you're having problems with readability, turn off the transparent background
    option so that all text is drawn on a clean background
      
  - All settings are saved in the vpp.ini file automatically
  
  - Options include the following:
  
      Font name, size and style
      Text colour (black or white only)
      Transparent background
      Automatic upper case
      Check clipboard for text
      Auto scaling (automatically matches text intensity to the target image)
      Alignment (the position of the text in the image)
      Text margin (a margin may be added at the edges of the text)
  
}
  
toolbar 'Text Overlay' ;

const
  { external library }
  TextOverlayLib = 'VTextOverlay.dll' ;

  { INI file identifiers }
  iniSection     = 'Text Overlay' ;
  iniFontName    = 'FontName' ;
  iniFontSize    = 'FontSize' ;
  iniFontStyle   = 'FontStyle' ;
  iniAlignment   = 'Alignment' ;
  iniMargin      = 'Margin' ;
  iniWhiteText   = 'WhiteText' ;
  iniTransparent = 'Transparent' ;
  iniForceCaps   = 'ForceCaps' ;
  iniAutoScale   = 'AutoScale' ;
  iniXFixed      = 'XFixed' ;
  iniYFixed      = 'YFixed' ;
  iniAutoPaste   = 'AutoPaste' ;
  
  { alignment options }
  alTopLeft     = 0 ;
  alTopRight    = 1 ;
  alBottomLeft  = 2 ;
  alBottomRight = 3 ;
  alFlag        = 4 ; // align to flag (top left) 
  alFixed       = 5 ; // align to a fixed coordinate (top left)

  { font style options }
  fsBold   = 1 ;
  fsItalic = 2 ;

  { text buffers }
  MaxFontNameSize = 128 ;
  
var
  { images }
  iOrig ;             // original image
  Rgn ;               // copy of unmodified region
  x1,y1,x2,y2 ;       // region of interest coordinates
  
  { text options }
  FFont,FSize ;       // font name and size
  FStyle ;            // font style
  FAlignment ;        // alignment option
  FMargin ;           // pixel margin to edge
  FWhiteText ;        // invert text to white (0/1)
  FTransparent ;      // transparent background (0/1)
  FForceCaps ;        // force text to upper case (0/1)
  FAutoScale ;        // scale maximum of text intensity to match image (0/1)
  FXFixed ;           // fixed alignment (x coord)
  FYFixed ;           // fixed alignment (y coord)
  FAutoPaste ;        // automatically preset the message from the clipboard (0/1)
  
  
{ External functions }  
function GetTextInfo( Msg,FontName:pointer; FontSize,FontStyle:integer; var tx,ty:integer ) : integer ; external TextOverlayLib ;
function CreateTextImage( Msg,FontName:pointer; FontSize,FontStyle:integer; Buffer:pointer ) : integer ; external TextOverlayLib ;
procedure ShowTextSetup( FontName:pointer; var nFontChars,FontSize,FontStyle,WhiteText,Transparent,ForceCaps,AutoPaste,AutoScale,Alignment,Margin,xf,yf:integer ) ; external TextOverlayLib ;


procedure GetBoundary( var Image,tx,ty,bx1,by1,bx2,by2 ) ;
{ Calculate the text region coordinates in the image }

var
  ix,iy ;
  Flags ;

begin
  { init }
  GetXYMax( Image,ix,iy ) ;

  { top left }
  if FAlignment = alTopLeft then begin
    bx1 := FMargin ;
    by1 := FMargin ;
  end;
  
  { top right }
  if FAlignment = alTopRight then begin
    bx1 := ( ix - FMargin ) - tx + 1 ;
    by1 := FMargin ;
  end;

  { bottom left }
  if FAlignment = alBottomLeft then begin
    bx1 := FMargin ;
    by1 := ( iy - FMargin ) - ty + 1 ;
  end;

  { bottom right }
  if FAlignment = alBottomRight then begin
    bx1 := ( ix - FMargin ) - tx + 1 ;
    by1 := ( iy - FMargin ) - ty + 1 ;
  end;

  { align to flag }
  if FAlignment = alFlag then begin
    Flags := GetFlags( Image ) ;
    if GetYSize( Flags ) >= 1 then begin
      bx1 := Flags[0,0] ;
      by1 := Flags[1,0] ;
    end
    else begin
      bx1 := FMargin ;
      by1 := FMargin ;
    end;
  end;
  
  { fixed alignment }
  if FAlignment = alFixed then begin
    bx1 := FXFixed ;
    by1 := FYFixed ;
  end;

  { calculate end coordinate }
  bx2 := bx1 + tx - 1 ;
  by2 := by1 + ty - 1 ;

end;  { GetBoundary }


function IsEnclosed( var Image,bx1,by1,bx2,by2 ) ;
{ Check text boundary is completely inside image }

var
  ix,iy ;

begin
  GetXYMax( Image,ix,iy ) ;
  IsEnclosed := ( ( bx1 >= 0 ) and ( by1 >= 0 ) and ( bx2 <= ix ) and ( by2 <= iy ) ) ;
end;  { IsEnclosed }


function TextAsImage( Msg,FontName,FontSize,FontStyle,iType ) ;
{ Create an image containing the message }

var
  Text,tx,ty ;
  Error ;
  
begin
  if GetTextInfo( Msg,FontName,FontSize,FontStyle,tx,ty ) = 0 then begin
    Text := CreateImage( typ_RGB,tx,ty ) ;
    if CreateTextImage( Msg,FontName,FontSize,FontStyle,Text ) = 0 then
      TextAsImage := Text ;   
  end;
end;  { TextAsImage }


function MaxScalar( a,b,c ) ;
{ Return the maximum of three scalar values }

begin
  if ( a >= b ) and ( a >= c ) then MaxScalar := a ;
  if ( b >= a ) and ( b >= c ) then MaxScalar := b ;
  if ( c >= a ) and ( c >= b ) then MaxScalar := c ;
end;  { MaxScalar }


function MaxComponent( var Image ) ;
{ Return the maximum component value of a color image }

begin
  MaxComponent := MaxScalar( MaxOf( Red( Image ) ),MaxOf( Green( Image ) ),MaxOf( Blue( Image ) ) ) ;
end;  { MaxComponent }


procedure ScaleText( var Image,iText ) ;
{ Scale the text image max intensity to match the target image }

var
  Factor ;
  
begin
  if ( FAutoScale = 1 ) then begin
  
    { calculate scaling factor } 
    if IsColor( Image ) then
      Factor := double( MaxComponent( Image ) ) / MaxComponent( iText )
    else
      Factor := double( MaxOf( Image ) ) / MaxOf( iText ) ;
      
    { scale text image }
    if Factor > 0 then begin
      ConvertType( iText,typ_double ) ;
      iText := iText * Factor ;  
      ConvertType( iText,TypeOf( Image ) ) ;
    end;
          
  end;      
end;  { ScaleText }


procedure InvertText( var Image,iText ) ;
{ Convert text to white on black }

begin
  if ( FWhiteText = 1 ) then iText := MaxOf( iText ) - iText ;
end;  { InvertText }


procedure BurnText( var Image,iText ) ;
{ Burn the text data into the image }

var
  iMask ;

begin
  { save area under text }
  Rgn := Image[ x1..x2,y1..y2 ] ;
  
  { process text }
  ConvertType( iText,TypeOf( Image ) ) ;
  InvertText( Image,iText ) ;
  ScaleText( Image,iText ) ;
  
  { copy to image }
  if ( FTransparent = 1 ) then begin
  
    { mask text }
    if ( FWhiteText = 1 ) then
      iMask := ( iText <> 0 )
    else
      iMask := ( iText <> MaxOf( iText ) ) ;

    { combine with image}
    Image[ x1..x2,y1..y2 ] := ( Rgn * not( iMask ) ) + ( iText * iMask ) ;
    
  end  
  else
    { copy text directly }
    Image[ x1..x2,y1..y2 ] := iText ;
    
  { update }
  Update( Image ) ;
          
end;  { BurnText }


procedure AddText ;
button btn_Text,'Add text|Overlay text onto an image' ;
menu '-Add &Text...' ;

var
  Image,iText,iMask ;
  tx,ty ;
  Msg ;
  
begin
  { get active image }
  GetActiveImage( iOrig ) ;
  if IsImage( iOrig ) then begin
  
    { get default message }
    if ( FAutoPaste = 1 ) then begin
      Msg := PasteFromClipboard ;
      if not IsString( Msg ) then Msg := '' ;
    end
    else
      Msg := '' ;
    
    { burn into image }
    if GetString( 'Enter text to burn into the image',Msg ) = id_Ok then 
      if Length( Msg ) > 0 then begin
      
        { force upper case ? }
        if ( FForceCaps = 1 ) then Msg := UpperCase( Msg ) ;
      
        { create text bitmaps }
        iText := TextAsImage( Msg,FFont,FSize,FStyle,TypeOf( iOrig ) ) ; 
        
        { insert text into image }
        if IsImage( iText ) then begin
          
          { check size and position }
          GetXYSize( iText,tx,ty ) ;
          GetBoundary( iOrig,tx,ty,x1,y1,x2,y2 ) ;
        
          { burn text into image }
          if IsEnclosed( iOrig,x1,y1,x2,y2 ) then 
            BurnText( iOrig,iText )
          else
            WriteError( 'Text overlaps image boundary' ) ;
        
        end
        else
          WriteError( 'Could not create text image' ) ;
            
      end;
      
  end;     
  
end;  { AddText }


procedure AddName ;
button btn_Text,'Add name|Overlay an image with its name' ;
menu '&Add Name...' ;

var
  OldAutoPaste ;
  
begin
  { get active image }
  GetActiveImage( iOrig ) ;
  if IsImage( iOrig ) then begin
  
    { turn on auto paste }
    OldAutoPaste := FAutoPaste ;
    FAutoPaste := 1 ;
    
    { annotate with file name }
    CopyToClipboard( GetName( iOrig ) ) ;
    AddText ;
    CopyToClipboard( '' ) ;
  
    { restore auto paste }  
    FAutoPaste := OldAutoPaste ;
     
  end;   
end;  { AddName }


procedure RemoveText ;
button btn_Text,'Remove|Remove text from the previous image' ;
menu '&Remove Text' ;

begin
  if IsImage( iOrig ) and IsImage( Rgn ) then
    if OnDesktop( iOrig ) then begin
      iOrig[ x1..x2,y1..y2 ] := Rgn ;
      Update( iOrig ) ;
    end;     
end;  { UndoText }


procedure CopyStringToBuffer( var Sz,Buffer ) ;
{ Copy a string to a memory buffer }

var
  i,n ;

begin
  { init }
  Buffer := 0 ;
  n := MaxScalar( Length( Sz ) + 1, GetXMax( Buffer ), 0 ) ;
  
  { copy } 
  for i := 0 to n - 1 do
    Buffer[i] := ord( ExtractStr( Sz,i+1,1 ) ) ;
      
end;  { CopyStringToBuffer }


procedure ReadSettings ;
{ Read settings from main INI file at startup }

begin
  FFont := ReadINIString( iniSection,iniFontName,'Arial' ) ;
  FSize := val( ReadINIString( iniSection,iniFontSize,'20' ) ) ;
  FStyle := val( ReadINIString( iniSection,iniFontStyle,'0' ) ) ;
  FAlignment := val( ReadINIString( iniSection,iniAlignment,'2' ) ) ;
  FMargin := val( ReadINIString( iniSection,iniMargin,'4' ) ) ;
  FWhiteText := val( ReadIniString( iniSection,iniWhiteText,'1' ) ) ;
  FTransparent := val( ReadIniString( iniSection,iniTransparent,'1' ) ) ;
  FForceCaps := val( ReadIniString( iniSection,iniForceCaps,'0' ) ) ;
  FAutoScale := val( ReadINIString( iniSection,iniAutoScale,'1' ) ) ;
  FXFixed := val( ReadINIString( iniSection,iniXFixed,'0' ) ) ;
  FYFixed := val( ReadINIString( iniSection,iniYFixed,'0' ) ) ;
  FAutoPaste := val( ReadINIString( iniSection,iniAutoPaste,'1' ) ) ;
end;  { ReadSettings }


procedure SaveSettings ; shutdown ;
{ Save settings to INI file at shutdown }

begin
  WriteINIString( iniSection,iniFontName,FFont ) ;
  WriteINIString( iniSection,iniFontSize,FSize ) ;
  WriteINIString( iniSection,iniFontStyle,FStyle ) ;
  WriteINIString( iniSection,iniAlignment,FAlignment ) ;
  WriteINIString( iniSection,iniMargin,FMargin ) ;
  WriteIniString( iniSection,iniWhiteText,FWhiteText ) ;
  WriteIniString( iniSection,iniTransparent,FTransparent ) ;
  WriteIniString( iniSection,iniForceCaps,FForceCaps ) ;
  WriteINIString( iniSection,iniAutoScale,FAutoScale ) ;
  WriteINIString( iniSection,iniXFixed,FXFixed ) ;
  WriteINIString( iniSection,iniYFixed,FYFixed ) ;
  WriteINIString( iniSection,iniAutoPaste,FAutoPaste ) ;
end;  { SaveSettings }


procedure Options ;
button btn_Text,'-Options|Set options for text overlay' ;
menu '&Options...' ;

var
  Buffer,MaxChars ;
  
begin
  { create externally accessible text buffer }
  Buffer := CreateArray( typ_Byte,MaxFontNameSize,1 ) ;
  CopyStringToBuffer( FFont,Buffer ) ;

  { show dialog box }
  MaxChars := GetXSize( Buffer ) ;  // needs to be a variable
  ShowTextSetup( Address( Buffer ),MaxChars,FSize,FStyle,FWhiteText,FTransparent,FForceCaps,FAutoPaste,FAutoScale,FAlignment,FMargin,FXFixed,FYFixed ) ;

  { process font name }
  FFont := BytesToStr( Buffer ) ;
  Free( Buffer ) ;
  
  { save new settings }
  SaveSettings ;
  
end;  { Options } 


begin
  { init }
  ReadSettings ;
  CopyToClipboard( '' ) ;
end.
